/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/




/*
* Modifications:
*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 * Modification author: Hejun Wu (vmnet@cs.ust.hk)
 
********************************************************************************/

#ifndef ATEMU_ATMEGA128_H
#define ATEMU_ATMEGA128_H

#include "defs.h"
#include "device.h"
#include "table.h"
#include "stack.h"

enum {
    ATEMU_MODE_DECODE,
		ATEMU_MODE_EXECUTE
};

typedef struct atmega_chip {
    int   id;
    char* xmlid;
	
    uint8_t *sram;
    uint32_t sram_bytes;
	
    uint32_t io_offset; /* offset to io space for in/out and friends */
    uint32_t io_last;   /* end of io space */
	
    uint16_t *flash;
    uint32_t flash_words;
	
    uint8_t pc_bits;    /* number of bits PC is */
    uint32_t pc_mask;   /* AND mask to bound flash access */
	
    /* docs say pc is 16 or 22 bits */
    uint32_t pc;
    uint32_t oldpc;
    uint32_t start_vector;
	
    uint8_t *sreg;                      /* Status Register */
    uint8_t sreg_io;                    /* SREG IO Register address */
	
#define ATEMU_SREG_I       0x80    /* Global Interrupt Enable */
#define ATEMU_SREG_T       0x40    /* Bit Copy Storage */
#define ATEMU_SREG_H       0x20    /* Half Carry Flag */
#define ATEMU_SREG_S       0x10    /* Sign Bit, S = N ^ V */
#define ATEMU_SREG_V       0x08    /* 2's Complement Overflow Flag */
#define ATEMU_SREG_N       0x04    /* Negative Flag */
#define ATEMU_SREG_Z       0x02    /* Zero Flag */
#define ATEMU_SREG_C       0x01    /* Carry Flag */
	
    uint8_t *register_file;            /* R0 - R31 */
	
    /* X, Y, and Z double-byte registers */
#define ATEMU_REG_X       0x1a
#define ATEMU_REG_X_LOW   0x1a
#define ATEMU_REG_X_HIGH  0x1b
#define ATEMU_REG_Y       0x1c
#define ATEMU_REG_Y_LOW   0x1c
#define ATEMU_REG_Y_HIGH  0x1d
#define ATEMU_REG_Z       0x1e
#define ATEMU_REG_Z_LOW   0x1e
	
#define ATEMU_REG_Z_HIGH  0x1f
	
    uint8_t *spl;
    uint8_t *sph;
	
    uint8_t spl_io;                     /* IO Space address of Stack pointer */
    uint8_t sph_io;                     /* IO Space address of Stack pointer */
	
    uint8_t *rampz;                     /* RAM Page */
    uint8_t rampz_io;                   /* IO Space address of rampz */
	
    uint8_t eind_io;                    /* IO Address of EIND */
    uint8_t *eind;
	
    uint64_t cycles;
    uint64_t oldcycles;
	
    /* internal registers */
    uint16_t instr;
    uint8_t th_nibble, tl_nibble, bh_nibble, bl_nibble, t_byte, b_byte;
	
    uint8_t reg_index_a;
    uint8_t reg_index_b;
    uint8_t a;
    uint8_t b;
    uint32_t addr;
	
    /* Bitmask of all interrupts (1 in bit position if active) */
    uint64_t interrupts;
	
    /* TRUE if the processor is currently executing a SLEEP op-code */
    int am_sleeping:1;
	
    /* io manager state */
    register_obj *register_head[256];
    interrupt_obj *interrupt_head;
    device_obj *device_head;
    clock_obj *clock_head;
	
    int mode;
	
    call_stack_t *call_stack;
    int stack_changed;
	
    table_t *breaktable;
    table_t *watchtable;
    table_t *notifytable;
    int watchhit;
	
    int sh;   /* stack height, used by the logger */
} atmega_chip;

typedef struct{
    uint32_t addr;
}break_t;

typedef struct{
    uint32_t addr;
    int length;
    int id;
    uint8_t *value;
}watch_t;

int breakcomp(const void *a, const void *b);
int watchcomp(const void *a, const void *b);

void atmega_branch_skip_instruction( atmega_chip *chip );

error_t atmega_read_sram( atmega_chip *chip, uint16_t addr,
						 uint8_t *value );
error_t atmega_write_sram( atmega_chip *chip, uint16_t addr,
						  uint8_t value );

int atemu_load_file( atmega_chip ***chips, char *fname, char *config );
error_t atmega_load_chip(atmega_chip **chips, int i, char *config, char *srec);
error_t atmega_create_chip( atmega_chip **ret_chip, char *conf );
error_t atmega_destroy_chip( atmega_chip **chip );
error_t atmega_load_flash( atmega_chip *chip, char *filename );
error_t atmega_step( atmega_chip *chip);

/* Cycle count to halt execution at */
extern uint64_t halttime;

//////////////////////////////////////////////////////////////////
//VMNet 
//Following are for logging of Emulation Manager
///////////////////////////////////////////////////////////////////////////

extern char g_isInitialized;
extern unsigned short g_idChip;
extern unsigned short g_location[];
extern atmega_chip ** g_chips;
extern unsigned long g_cycles[];
extern unsigned long g_oldcycles[];
extern unsigned char g_flagSleep[];
uint64_t get_program_time (struct timeval *tv);

//defination
#define MODE_SLEEP 0x0
#define MODE_PROCESS 0x1
#define MODE_TRANSMIT 0x2
#define MODE_RECEIVE  0x3
#define ACTION_NORMAL 0
#define SET_ACTION_ADC    0x1
#define SET_ACTION_FLASH  0x2
#define CLEAR_ACTION_ADC    0xe
#define CLEAR_ACTION_FLASH  0xd

#define DEBUG_VERSION
#ifdef DEBUG_VERSION
#define LOG_CURRENT_FUNCTION(strName, locationBranch)if (logfile)\
{\
	fprintf(logfile, "%03d %s: %d \t\n", g_idChip,strName,locationBranch);\
}
#else
#define LOG_CURRENT_FUNCTION(strName, locationBranch)
//   printf( "%03d %s: %x \t\n", g_idChip,strName,locationBranch);
#endif

#define DEBUG_OUTPUT
#define DEBUG_TIMER
#define DEBUG_USART printf
#define DEBUG_USART_FUNCTION(strName, locationBranch)  printf( "%02d %s: %x \t\n", g_idChip,strName,locationBranch&0xff);

extern char g_debugFlag;
extern int g_tempDebug;
extern int g_num_attributes;
//////////////////////////////////////////////////////////////////////////////////////////////
// VMNet
// Following are for operations of components in a sensor node,
// which are important for power evaluation
//////////////////////////////////////////////////////////////////////////////////////////////

#define CPU_OPERATION_TYPES    3
#define MAX_NUM_LOGS 100
#define MAX_OPERATION_STR_LEN 20
extern unsigned char strCPUOperations[CPU_OPERATION_TYPES][MAX_OPERATION_STR_LEN];
extern char status_cpu[MAX_NUM_LOGS];
extern uint16_t g_status_log[MAX_NUM_LOGS];


#define CPU_SLEEP 0x0100
#define CPU_ACTIVE  0x0200

#define ADC_SLEEP  0x0010
#define ADC_CONTROL  0x0020
#define ADC_SAMPLE  0x0040
#define ADC_IDLE    0x0080

#define RADIO_SLEEP 0x0001
#define RADIO_IDLE  0x0002
#define RADIO_CONTROL 0x0000
#define RADIO_RECEIVE  0x0004
#define RADIO_TRANSMIT  0x0008

#define CPU_MASK  0x00ff
#define ADC_MASK  0x0f0f
#define RADIO_MASK 0x0ff0
#define MAX_SENSOR_DATA_SIZE 1000
#define MAX_SENSOR_ATTRIBUTES 3


#define POWER_CONSUMPTION_MEASURE 1

#define LOG_CONTROL_START 0 // todo: define the start epoch
#define LOG_CONTROL_END 1// todo: define the end epoch.
#define NORMAL_LOG
#define EXP_TIMES 30
#define MOTE_ID_LOCATION 36
#define REFERENCE_MOTE_ID 1
#define QUERY_RESULT_FLAG 0xf0
#define QUERY_RESULT_LOCATION 5

extern char g_log_controllreg;


#ifdef POWER_CONSUMPTION_MEASURE

#define FIXED_MEASURE
extern char g_log_changed_flag[MAX_NUM_LOGS] ;
extern   char g_logIDMote;

#define LOG_CURRENT_OPERATION(operationtype,mask) {  if(g_power_log[g_idChip])\
{\
	if((uint16_t)(operationtype|mask)!=(uint16_t)(g_status_log[g_idChip]|mask))\
{\
				g_status_log[g_idChip]&=mask;\
				g_status_log[g_idChip]|=operationtype;\
				g_oldcycles[g_idChip] = g_chips[g_idChip]->cycles;\
				if((g_log_controllreg<LOG_CONTROL_END)&&(g_log_controllreg>=LOG_CONTROL_START))\
{fprintf(g_power_log[g_idChip], "%x %lu \t\n",g_status_log[g_idChip],g_chips[g_idChip]->cycles); };\
}\
}\
}
#define DUMP_POWER_LOG() if((g_logIDMote==-1)||(g_logIDMote==g_idChip)) {if(g_log_changed_flag[g_idChip]) {fprintf(g_power_log[g_idChip], "%d %d \t\n",g_status_log[g_idChip],g_chips[g_idChip]->cycles); g_log_changed_flag[g_idChip]=0;};}

#else

#define LOG_CURRENT_OPERATION(operationtype,mask)     
#define LOG_CURRENT_DEVICE_MODE(strName,mode) 

#endif //POWER_CONSUMPTION_MEASURE

//////////////////////////////////////////////////////////////////////////////////////////////
//VMNet
//Following are for multi-PCs.
/////////////////////////////////////////////////////////////////////////////////////

#define COMMENCE_SYCHRONIZATION 10
#define VMNET_BCAST_ADDR 0xffff
#define STANDARD_LENGTH  500
#define BIT_PORT 10600
#define NET_MANAGER_PORT 5000
#define BIT_EVENT_PORT 10701
#define MAX_MOTE_PORTS 3
#define RFM_BIT_TRANSMIT 100
#define localAddr INADDR_LOOPBACK
#define	BROADCAST_ADR "192.168.1.255"
#define MAX_MOTES 1000
#define STABLE_EXECUTE_TIMES  1000

#define OPERATION_OPTIMIZE  1

extern int g_idVMNet;
extern int g_numNodes;
void initializeRadioSockets(int port) ;
int writeRadioMessages(void *data, int datalen, int destAddr);
int readRadioMessages(void *data, int datalen, int timeout);
int get_id_synchronization(char* data);
int Synchronization(short int id);
int IsNeedSynchronization(atmega_chip * chips, short int id);
/////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////



#endif /* ATEMU_ATMEGA128_H */
