/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/    

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 * Modification author: Hejun Wu (vmnet@cs.ust.hk)
 
********************************************************************************/




#include <sys/stat.h>
#include <fcntl.h>
#include "defs.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <stdio.h>

#include "srec.h"
#include "atmega128.h"
#include "iomanager.h"
#include "debug.h"
#include "parse_config.h"
#include "table.h"


static char sreg_flags[9] = "CZNVSHTI";
int interrupt=0;
static int currchip=0; 
static int nchips=0;   
//VMNet for EM logs+++++++++++++++

//++++++++++++++++++++++++++++++++
static int backend = 0;
int verbosity=V_NORMAL;
uint64_t currentcycle = 0;

FILE *logfile=NULL;

int logstack=0;
int logmachine=0;
int logmemory=0;
int logcall=0;
int logintr=0;
int logtask=0;

///////////////////////////////////
//VMNet
//The following are for performance logging
//////////////////////////
FILE **g_power_log;
char g_logIDMote=-1;
uint16_t g_status_log[MAX_NUM_LOGS]={0};
uint16_t *g_position_x,*g_position_y,*g_position_z;
//int32_t *g_current_waited_cycles;
#ifdef LOG_INITIALIZATION_TIME
uint8_t *g_received_id;
float g_received_percentage=0.6;
uint16_t g_received_count=0;
uint32_t *g_epoch_id;
uint32_t g_total_results=0;
FILE* g_epoch_result_log;
#endif
char g_log_changed_flag[MAX_NUM_LOGS]={0};
FILE **g_sensor_data_fptr;
long *g_sample_data_count;
int g_num_attributes=1;
unsigned short **g_sensor_data;
#ifdef FIXED_MEASURE
char g_log_controllreg = -1;//whj
#else
char g_log_controllreg = 3;//whj
#endif                      
int g_nChips;
uint64_t get_program_time (struct timeval *tv)
{
    uint64_t result;
    if (gettimeofday (tv, NULL) < 0)
        printf ("Failed to get program time.\n");
    result = ((uint64_t) tv->tv_sec * 1000) + ((uint64_t) tv->tv_usec / 1000);
    return result;
}

////////////////////////////////////////////////////////
//VMNet: test for multiple PCs. not finished yet...
//
////////////////////////////

char g_isInitialized=0;
//later use whj.
char IP_ADDRESSES[3][16]={ "192.168.1.7","192.168.1.15", "192.168.1.14"};
unsigned long g_cycles[MAX_NUM_LOGS]={0L};
unsigned long g_oldcycles[MAX_NUM_LOGS]={0L};
unsigned char g_flagSleep[MAX_NUM_LOGS]={0};
int g_fd;
struct timeval startTime, stopTime;
uint64_t startValue, stopValue;
atmega_chip ** g_chips;
/////////////////
int g_idVMNet;
int g_numNodes=0;

/////////////////////////////////////////////////////////////////////////////
int bitServerSocket = -1;
//int bitEventServerSocket = -1;
int bitClientSocket=-1;
char g_synchronized[MAX_MOTES]={0};
void initializeRadioSockets(int port) {
	int i;
	//whj -------------2004-1-4 11:25---------------------------------------
	//-----------------------------------------------------------------------
    struct sockaddr_in sock;
	int sfd,enable;
	int rval = -1;
    //whj -------------2004-1-4 11:25---------------------------------------
	//-----------------------------------------------------------------------
	//  printf("PORT is %d \n", port);
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_port = htons(port); 
	sock.sin_addr.s_addr = htonl(INADDR_ANY);
	
	bitServerSocket = socket(AF_INET, SOCK_DGRAM, 0);
	//----------------------------------------------------------------------------------
	//use udp SOCK_DGRAM not SOCK_STREAM
	sfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (port<(BIT_PORT+NET_MANAGER_PORT))
	{
		
		int n=  fcntl(sfd,F_GETFL);
		//    printf("Set, hahhahh\n");
		fcntl(sfd, F_SETFL, n|O_NONBLOCK);
    }
	
	if (sfd < 0) {
		printf( "ERROR: Could not create server socket: %s\n", strerror(errno));
		exit(-1);
	}
	setsockopt(sfd, SOL_SOCKET, SO_BROADCAST , (char *)&enable, sizeof(int));
	while(rval < 0) {
		rval = bind(sfd, (struct sockaddr*)&sock, sizeof(sock));
		if (rval < 0) {
			printf( "ERROR: Could not create server socket: %s\n", strerror(errno));
			exit(-1); 
		}
	}
	bitClientSocket = sfd;
	
}

int writeRadioMessages(void *data, int datalen, int destAddr) {
	unsigned char ack;
	int i=0, j;
	short nMaxPorts;
	struct sockaddr_in sock, remotesock;
	sock.sin_family = AF_INET;
	sock.sin_port = htons(BIT_PORT);
	sock.sin_addr.s_addr = inet_addr("127.0.0.1");
	remotesock.sin_family = AF_INET;
	remotesock.sin_port = htons(BIT_PORT);
	
	if (destAddr==VMNET_BCAST_ADDR)
	{
		nMaxPorts=MAX_MOTE_PORTS;
		for(j = 0;j<nMaxPorts;j++)
		{
			remotesock.sin_addr.s_addr = inet_addr(IP_ADDRESSES[j]);
			sock.sin_port = htons(BIT_PORT+j);
			remotesock.sin_port = htons(BIT_PORT+j);
			i =sendto(bitServerSocket,data,datalen,0,(struct sockaddr *)&remotesock,sizeof(remotesock));
			// i =sendto(bitServerSocket,data,datalen,0,(struct sockaddr *)&sock,sizeof(sock));
			
			if (i < 0)
			{
				
				fprintf(stderr, "ERROR: Socket closed: %s\n", strerror(errno));
			}
		}
	}
	else
	{
		if(destAddr!=NET_MANAGER_PORT)
			remotesock.sin_addr.s_addr = inet_addr(IP_ADDRESSES[j]);
		else { remotesock.sin_addr.s_addr = inet_addr(IP_ADDRESSES[0]);}
		sock.sin_port = htons(BIT_PORT+destAddr);
		remotesock.sin_port = htons(BIT_PORT+destAddr);
		i =sendto(bitServerSocket,data,datalen,0,(struct sockaddr *)&remotesock,sizeof(remotesock));
		//i =sendto(bitServerSocket,data,datalen,0,(struct sockaddr *)&sock,sizeof(sock));
		if (i < 0)
		{
			
			fprintf(stderr, "ERROR: %s\n", strerror(errno));
		}
	}
	return 0;
}


int readRadioMessages(void *data, int datalen, int timeout)
{
	struct sockaddr_in sock;
	int i;
	int addrlen;
	sock.sin_family = AF_INET;
	sock.sin_port = htons(BIT_PORT);
	sock.sin_addr.s_addr = inet_addr("127.0.0.1");
	memset((char*)&sock,0,sizeof(sock));
	addrlen=sizeof(sock);
	if(bitClientSocket==-1)
	{
		printf("ERROR: no initialized socket \n");
		exit(-1);
	}
	if (timeout)
		i =recvfrom(bitClientSocket,data,datalen,MSG_DONTWAIT,( struct sockaddr*)&sock,&addrlen);
	else
		i =recvfrom(bitClientSocket,data,datalen,0,( struct sockaddr*)&sock,&addrlen);
	
	return i;
}
int get_id_synchronization(char* data)
{
	int id;
	if((char)(data[0])!=(char)(0xff))
		return -1;
	id=data[4];
	
	id<<=8;
	// id&=0xff00;
	return (id|data[5]);
}

short int get_action_synchronization(char* data)
{
	short int id;
	if((char)(data[0])!=(char)(0xff))
		return -1;
	id=data[6];
	id<<=8;
	return (id|data[7]);
}
int SetCommenceData(char * data)
{
	int id;
	data[6]=0;
	data[7]=COMMENCE_SYCHRONIZATION;
	return 0; 
}
int RequestSynchronization(short int id)
{
	int i=0, meaning=0;
	char synchronizationData[16],receivedData[16];
	for(i=0;i<4;i++)
		synchronizationData[i]=0xff;
	synchronizationData[4]=(0xff00&id)>>8;
	synchronizationData[5]=(0xff&id);
	i=0;
	//  printf("Request synchronization.....\n");
	SetCommenceData(synchronizationData);
	while(!interrupt)
	{
		writeRadioMessages(synchronizationData, 16,NET_MANAGER_PORT);
		readRadioMessages(receivedData, (int)16, 10);
		meaning=get_action_synchronization(receivedData);
		id=get_id_synchronization(receivedData);
		if((id<0)&&(meaning==COMMENCE_SYCHRONIZATION))
		{
			//         printf("Request accept !!!\n");
			return 1;
		}
		
	}
	return 0;
	
}
int Synchronization(short int id)
{
	int i=0,getID, meaning=0;
	static char firsttime=1;
	char synchronizationData[16],receivedData[16];
	memset(synchronizationData,0,16);
    memset(receivedData,0,16);
	for(i=0;i<4;i++)
		synchronizationData[i]=0xff;
	//       printf("prepare synchronization  %d.....\n", id);
	synchronizationData[4]=(((unsigned short int)(0xff00&id))>>8);
	synchronizationData[5]=(0xff&id);
	
	id=get_id_synchronization(synchronizationData);
	//   printf("Begin synchronization  %d.....\n", id);
	//   printf("interrupt====%d\n",interrupt);
	memset( g_synchronized,0, MAX_MOTES);
	i=0;
	while(!interrupt)
	{
		
		if(id<0)
		{
			
			{
				readRadioMessages(receivedData, (int)16,0);
				getID=get_id_synchronization(receivedData);
				if(getID<0)continue;
				meaning=get_action_synchronization(receivedData);
				if(meaning==COMMENCE_SYCHRONIZATION)
				{ int j;
				//         printf("From ID:   %d\n", getID);
				SetCommenceData(synchronizationData);
				for(j=0;j<COMMENCE_SYCHRONIZATION;j++)
				{
					writeRadioMessages(synchronizationData, 16, VMNET_BCAST_ADDR);
					usleep(10);
				}
				continue;
				}
				if(!(g_synchronized[getID]))
				{
					// printf("Received msg from: %d\n",getID);
					i++;
					g_synchronized[getID]=1;
				}
				//     printf("Received msg from: %d\n",getID);  
				
				if(i<g_numNodes)
					continue;
				
			}
			
			
			writeRadioMessages(synchronizationData, 16, VMNET_BCAST_ADDR);
			return;
		}
		else
		{
			
			writeRadioMessages(synchronizationData, 16,NET_MANAGER_PORT);
			usleep(100);
			getID=readRadioMessages(receivedData, 16,100);
			//  printf("synchronization ...\n");
			if(getID<0)
			{
				usleep(10);
				
				continue;
			}
			getID=get_id_synchronization(receivedData);
			meaning=get_action_synchronization(receivedData);
			if((getID<0)&&(meaning==COMMENCE_SYCHRONIZATION))
				continue;
			if(getID==-2)
			{
				// printf("synchronization done\n");
				if(firsttime)
					startValue=get_program_time(&startTime);
				firsttime=0;
				return;
			}
			
		}
	}
}

int IsNeedSynchronization(atmega_chip * chip, short int id)
{
	int i=0,getID, ret=0, meaning=0;
	static char firstTime=1;
	char receivedData[16];
	return 1;
	if (firstTime )
	{
		firstTime=0;
		return 1;
	}
	if(chip->cycles%(3000000L))return 0;
	return 1;
	getID=readRadioMessages(receivedData, 16,100);
	if(getID<0)
		ret= 0;
	getID=get_id_synchronization(receivedData);
	if(getID>=0) ret=0;
	else
	{
		meaning=get_action_synchronization(receivedData);
		if(getID<0&&meaning==COMMENCE_SYCHRONIZATION)
			return 1;
	}
	
    
	return 0;
	
}
//Finished VMNet test.
////////////////////////////////////////////////////////////////////////////////////////////

static void signal_handler(int type){
    signal(type, SIG_IGN);
    
    switch(type){
	case SIGINT:
		interrupt=1;
		printf("quit....\n");
		exit(1);
	case SIGCHLD:
		wait(NULL);
		//            printf("signal chiled\n");
		break;
	default:
		printf("signal handler received unknown signal\n");
		break;
    }
    signal(type, signal_handler);
}

static void execute_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    for(i=0; i<nchips; i++){
        /* execute current */
        g_idChip=i;
        if(chips[i]->cycles == currentcycle) {
            chips[i]->mode=ATEMU_MODE_EXECUTE;
            /* If time has been frozen, execute chip until no longer frozen */
            if(chips[i]->sram[0x00FA]) {
                uint64_t orgcycle = chips[i]->cycles;
                while(chips[i]->sram[0x00FA]) {
                    atmega_step(chips[i]);
                }
                chips[i]->cycles = orgcycle + 1;
            } else {
                atmega_step( chips[i]);
                /* break on unimplemented instructions */
                if(chips[i]->cycles == currentcycle){
                    interrupt = 1;
                }
            }
        }
        atmega_ioman_tick_clocks(chips[i]);
    }
}

static void decode_chips(atmega_chip **chips, uint64_t currentcycle){
    int i;
    
    for(i=0; i<nchips; i++){
        /* try and show where we will be, come next clock tick */
        if(chips[i]->am_sleeping){
            /* we are sleeping, assume we will continue */
            chips[i]->mode=ATEMU_MODE_DECODE;
            chips[i]->pc--;
            atmega_step( chips[i] );
            chips[i]->pc++;
        }else if(chips[i]->cycles == currentcycle){
            /* we broke right before the gicen instruction */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );
        }else{
		/* we are currently executing, and will continue to
			* execute this in the next cycle */
            uint32_t pc;
            uint64_t cycles;
			
            /* save values */
            pc=chips[i]->pc;
            cycles=chips[i]->cycles;
			
            /* move into the past */
            chips[i]->pc=chips[i]->oldpc;
            chips[i]->cycles=chips[i]->oldcycles;
			
            /* decode */
            chips[i]->mode=ATEMU_MODE_DECODE;
            atmega_step( chips[i] );
			
            /* restore values */
            chips[i]->pc=pc;
            chips[i]->cycles=cycles;
        }
    }
}

static int fork_compress (const char *fn)
{
    int fds[2], fd;
    int ires[2];
    unsigned int res;
    FILE *f;
	
    // hack -- better to discover problems before forking
    fd = open(fn, O_CREAT | O_WRONLY);
    if (fd == -1) {
        printf("couldn't open or create output file\n");
        exit(-1);
    }
	
    res = pipe (fds);
    if (res != 0) {
        printf("Output compress pipe failed.\n");
		return -1;
    }
	
    if (fork() != 0) {
        // parent -- close the write side and return the read side
        close (fds[0]);
		g_fd = fds[1];
		return fds[1];
    }
	
    // child -- copy the write side over STDIN, file to STDOUT
    ires[0] = dup2(fds[0], 0);
    ires[1] = dup2(fd, 1);
    if (ires[0] == -1 || ires[1] == -1) {
		printf("dup2 failed\n");
		exit (-1);
    }
	
    // make sure signal handlers are off for this child process
    signal(SIGINT, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    close (fds[0]);
    close (fds[1]);
	
    // never returns
    res = execlp ("bzip2", "bzip2", "-9", NULL, NULL);
	
    printf("bzip2 exec failed\n");
    exit (-1);
}

static void dump( atmega_chip *chip, uint8_t *mem, uint16_t length, uint16_t begin )
{
    int addr, inchar;
    char memchar;
	
    mem += begin;
	
    for( addr = 0; addr < length; addr += 16 )
    {
        /* so I sort of cheat... */
        log_dump(chip, "");
		
        printf( "%06x:    ", addr + begin );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x  ",
            mem[addr], mem[addr + 1], mem[addr + 2], mem[addr + 3],
            mem[addr + 4], mem[addr + 5], mem[addr + 6], mem[addr + 7] );
        printf( "%02x %02x %02x %02x %02x %02x %02x %02x",
            mem[addr + 8], mem[addr + 9], mem[addr + 10], mem[addr + 11],
            mem[addr + 12], mem[addr + 13], mem[addr + 14], mem[addr + 15] );
        printf("    ");
        for(inchar = 0; inchar < 16; inchar++) {
            memchar = mem[addr + inchar];
            putchar(((memchar < 32) || (memchar > 126)) ? '.' : memchar);
        }
        printf("\n");
    }
	
    fflush(stdout);
}

static void unassemble( atmega_chip *chip, int id, char *file ){
    uint32_t oldpc;
    int delta;
    int nnops=0;
    
    oldpc=chip->pc;
    chip->pc=0;
    chip->mode=ATEMU_MODE_DECODE;
    set_decode_mode(DECODE_DUMP, file);
	
    do{
	/* cheat around the fetch stage, to avoid having to print
		* thousands of NOPs at the end of the program */
        if(chip->flash[chip->pc]==0x0000){
            nnops++;
            delta=1;
            chip->pc++;
        }else{
            if(nnops>0){
                int i;
                chip->pc-=nnops;
                for(i=0; i<nnops; i++){
                    log_opcode(chip, "NOP\n");
                    chip->pc++;
                }
                nnops=0;
            }
            delta=atmega_step(chip);
            if(delta == -1){
                /* illegal instruction, do it as bytes */
                log_opcode(chip, ".byte\t0x%02x, 0x%02x\n", BOTTOM_BYTE(chip->flash[chip->pc]), TOP_BYTE(chip->flash[chip->pc]));
                delta=1;
            }
            chip->pc+=delta;
        }
    }while(chip->pc < chip->flash_words && (delta == 1 || delta == 2) && !interrupt);
	
    chip->pc=oldpc;
    set_decode_mode(DECODE_NORMAL, NULL);
}

static void dump_regs( atmega_chip *chip )
{
    char sreg[9];
    int i;
    int reg;
	
    for ( i = 0; i < 8; i++ )
        sreg[i] = ( GET_FLAG( *chip->sreg, 1 << i ) ) ? 
		(sreg_flags[i]) : (' ');
    sreg[8] = 0;
	
    printf( "[%lld] PC: 0x%08x\tSREG: 0x%02x[%s]\tSTACK: 0x%02x%02x\n",
        chip->cycles, chip->pc * 2, *chip->sreg, sreg, *chip->sph, *chip->spl );
	
    for( reg = 0; reg < 32; reg++ )
    {
        printf( "r%2d: %02x  ", reg, chip->register_file[reg] );
		
        if (! ( (reg + 1) % 8 ) )  printf( "\n" );
    }
	
    printf( "\n\n" );
}



static void dump_flash( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->flash, dump_bytes, dump_addr );
}

static void dump_sram( atmega_chip *chip, uint32_t dump_addr , uint32_t dump_bytes)
{
    dump( chip, (uint8_t*)chip->sram, dump_bytes, dump_addr );
}

static void has_changed_npoint( void *a, void *extra )
{
    watch_t *w = (watch_t *) a;
    atmega_chip *chip = (atmega_chip *) extra;
    uint8_t curr_val;
	
    atmega_read_sram(chip, w->addr, &(curr_val));
	
    if ( curr_val != w->value[0] )
    {
        w->value[0] = curr_val;
		
        log_debug(chip, V_QUIET,
			"notify addr %x id %d = %x, chip %d\n",
			w->addr, w->id, w->value[0], chip->id);
        fflush( stdout );
    }
}

static void mainloop_step(atmega_chip **chips) {
    int i;
	
    /* run current instruction */
    execute_chips(chips, currentcycle);
    currentcycle++;
	
    /* check notify points */
    for(i = 0; i < nchips; i++) {
        table_foreach(chips[i]->notifytable, has_changed_npoint, chips[i]);
    }
	
    for(i = 0; i < nchips; i++) {
        if (chips[i]->stack_changed != 0) {
            chips[i]->stack_changed = 0;
            stack_dump(chips[i]->pc, chips[i]->call_stack, chips[i]->id);
        }
    }
	
    /* decode next instruction */
    decode_chips(chips, currentcycle);
}

static void mainloop_go(atmega_chip **chips, int breakset, int watchset) {
    int i;
    int breakhit = 0; /* have we hit a brealpoint? */
	
    /* clear interrupt flag, if user pressed ctrl-c */
    do{
        execute_chips(chips, currentcycle);
        currentcycle++;
		//+++++++++++++++++++++++++++++++++
		//Currently, VMNet modifies the function to run the instructions without halt and break...
		//++++++++++++++++++++++++++++++++++
        if(breakset > 0 || watchset > 0) {
            break_t *bpoint, query;
            for(i=0; i < nchips; i++) {
                if(chips[i]->cycles == currentcycle && !chips[i]->am_sleeping){
                    query.addr = chips[i]->pc;
                    bpoint = table_find(chips[i]->breaktable, &query);
                    if(bpoint != NULL){
                        breakhit++;
                        log_debug(chips[i], V_QUIET,
							"breakpoint 0x%06x hit.\n", 2 * bpoint->addr);
                    }
					
                    /* trigger a "breakpoint" if a watch was hit */
                    if(chips[i]->watchhit != 0){
                        chips[i]->watchhit = 0;
                        breakhit++;
                    }
                }
            } /* for(i=0; i < nchips; i++) { */
        } /* if(breakset > 0 || watchset > 0) { */
		
    } while (!interrupt && !breakhit);
	
    decode_chips(chips, currentcycle);
	
    for(i=0; i < nchips; i++) {
        table_foreach( chips[i]->notifytable,
			has_changed_npoint, chips[i] );
        //log_header(chips[i],"END")
        stack_dump( chips[i]->pc, chips[i]->call_stack, chips[i]->id );
    }
    
}

static void mainloop(atmega_chip **chips) {
    int quit = 0;
    /* number of current *points active */
    int breakset = 0;
    int watchset = 0;
    int i;
    char input[1024];
    /* variables utalized during parsing */
    char *parse, *lastchar;
    char **argv = NULL;
    int  argc, argv_alloc = 0;
    /* variables used in decoding parameters */
    int  intparse;
    uint32_t addrparse, addrbytes;
    break_t *bpoint, bquery;
    watch_t *wpoint, wquery;
    static char flagRun=0;
    /* Make initial decodes upon starting program */
    for(i = 0; i < nchips; i++){
        chips[i]->mode = ATEMU_MODE_DECODE;
        atmega_step(chips[i]);
        stack_dump( chips[i]->pc, chips[i]->call_stack, chips[i]->id );
    }
	
    /* Prepare for parsing */
    argv_alloc = 8;
    argv = (char**)malloc(argv_alloc * sizeof(char*));
	
    while( !quit )
    {
        
        if(flagRun)
        {
			
			if (fgets(input, sizeof(input), stdin)) 
			{
				// Break command into argv/argc style /
				argc = 0;
				parse = strtok(input, " \t\n\r");
				
				while(parse != NULL) {
					// Make sure we have space to store this argument
					if(argv_alloc == argc) {
						argv_alloc += 8;
						argv = realloc(argv, argv_alloc * sizeof(char*));
						if(argv == NULL) {
							perror("realloc()");
							exit(-1);
						}
					}
					
					// Store this argument /
					argv[argc++] = parse;
					
					parse = strtok(NULL, " \t\n\r");
				}
			}
			else  {  //if ( fgets( input, sizeof(input), stdin) ) 
				if(feof(stdin)){
					quit = 1;
				}else{
					// if they pressed ^C during fgets, forget it /
					interrupt = 0;
				};
            }
			
		}
		// If argc is 0, a blank line was entered and we treat it as
		// * a step /
		
		if(argc == 0) {
			mainloop_step(chips);
			continue;
		}
		{
			
			char argg='G';
            
			if(flagRun)
				argg=toupper(argv[0][0]);
			flagRun=1;
            /* Handle specified command */
            switch(argg) {
            case 'Q':
                /* Quitting the program */
                quit = 1;
                break;
				
            case '?':
                /* Getting the command summary from the console */
                printf("Commands (case-insensitive):\n");
                printf("\tS or Enter \t\tStep one instruction\n");
                printf("\tD address [length]\tDump the SRAM\n");
                printf("\tF address [length]\tFlash dump\n");
                printf("\tR\t\t\tRegister dump\n");
                printf("\tU [output file]\t\tUnassemble program\n");
                printf("\tC chip number\t\tChip select\n");
                printf("\tB address\t\tBreakpoint at the Hex Address\n");
                printf("\tW address\t\tWatchpoint at the Hex Address\n");
                printf("\tN address\t\tNotfiypoint at the Hex Address\n"); 
                printf("\tX B|W|N addr\t\tDelete Break, Watch or Notify point at addr\n");
                printf("\tJ address\t\tJumps to the Hex Address\n");
                printf("\tG\t\t\tGo until breakpoint\n");
                printf("\tQ\t\t\tQuit\n");
                break;
				
            case 'C':
                /* Changing the active chip */
				
                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
						" argument.\n", 'C');
                    continue;
                }
				
                /* Read the argument */
                intparse = strtoul(argv[1], &lastchar, 10);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" decimal number.\n", argv[1]);
                    continue;
                }
				
                /* Parameter doesn't contain any illegal characters */
                if((intparse < 0) || (intparse >= nchips)) {
                    /* Error: parameter out of range */
                    printf("Error: The value %d is out of range"
						" (chips are numbered 0 through %d).\n",
						intparse, nchips-1);
                    continue;
                }
				
                /* Change the current chip */
                currchip = intparse;
                printf("Current chip is now %d.\n", currchip);
                break;
				
            case 'S':
                /* Single stepping */
                mainloop_step(chips);
                break;
				
            case 'F':
                /* Dump bytes from flash */
				
                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires at least the"
						" starting address to be provided.\n", 'F');
                    continue;
                }
				
                /* Parse the address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                /* Round to the nearest 16 */
                addrparse = (addrparse / 16) * 16;
				
                /* Make sure it was a valid hexadecimal number */
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Illegal character */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                if(argc > 2) {
                    /* Parse the byte count */
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    /* Make sure it was a valid hexadecimal number */
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Illegal character */
                        printf("Error: The value '%s' is not a valid"
							" hexadecimal number.\n", argv[2]);
                        continue;
                    }
                } else {
                    /* Default to 16 bytes if none provided */
                    addrbytes = 16;
                }
				
                /* Bound requested range to memory size */
                if(addrparse >= (chips[currchip]->flash_words * 2)) {
                    /* Beginning of range is after top of memory */
                    continue;
                }
                if((addrparse + addrbytes) >
					(chips[currchip]->flash_words * 2)) {
                    /* End of range is after top of memory */
                    addrbytes = (chips[currchip]->flash_words * 2) - addrparse;
                }
				
                dump_flash(chips[currchip], addrparse, addrbytes);
                break;
				
            case 'J':
                /* Set's the program counter */
				
                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
						" argument.\n", 'J');
                    continue;
                }
				
                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                /* Make sure the address is in range */
                addrparse /= 2;
                if(addrparse >= chips[currchip]->flash_words) {
                    printf("Error: The address '%s' is out of range"
						" (addresses limited to 0 - 0x%X).\n",
						argv[1], chips[currchip]->flash_words * 2 - 1);
                    continue;
                }
				
                chips[currchip]->pc = addrparse;
                chips[currchip]->mode = ATEMU_MODE_DECODE;
                atmega_step(chips[currchip]);
                break;
				
            case 'P':
                /* Poking holes in SRAM */
				
                /* Check arguments */
                if(argc < 3) {
                    printf("Error: The P command requires both an address"
						" and a byte value to be provided.\n");
                    continue;
                }
				
                /* Parse the arguments */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
                addrbytes = strtoul(argv[2], &lastchar, 0);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
						" numerical argument.\n", argv[2]);
                    continue;
                }
				
                /* Make sure the byte value is in range */
                if((addrbytes < -128) || (addrbytes > 255)) {
                    printf("Error: The value '%s' is larger than"
						" one byte.\n", argv[2]);
                    continue;
                }
				
                /* Make sure the address is within range */
                if(addrparse >= chips[currchip]->sram_bytes) {
                    printf("Error: The address '%s' is out of range"
						" (addresses limited to 0 - 0x%X).\n",
						argv[1], chips[currchip]->sram_bytes - 1);
                    continue;
                }
				
                chips[currchip]->sram[addrparse] = addrbytes;
                break;
				
            case 'R':
                /* Dump the CPU registers in a human-readable manner */
                dump_regs(chips[currchip]);
                break;
				
            case 'D':
                /* Dump bytes from SRAM */
				
                /* Check arguments */
                if(argc < 2) {
                    printf("Error: The %c command requires at least the"
						" starting address to be provided.\n", 'D');
                    continue;
                }
				
                /* Parse the address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                /* Round to the nearest 16 */
                addrparse = (addrparse / 16) * 16;
				
                /* Make sure it was a valid hexadecimal number */
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Illegal character */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                if(argc > 2) {
                    /* Parse the byte count */
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    /* Make sure it was a valid hexadecimal number */
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Illegal character */
                        printf("Error: The value '%s' is not a valid"
							" hexadecimal number.\n", argv[2]);
                        continue;
                    }
                } else {
                    /* Default to 16 bytes if none provided */
                    addrbytes = 16;
                }
				
                /* Bound requested range to memory size */
                if(addrparse >= chips[currchip]->sram_bytes) {
                    /* Beginning of range is after top of memory */
                    continue;
                }
                if((addrparse + addrbytes) > chips[currchip]->sram_bytes) {
                    /* End of range is after top of memory */
                    addrbytes = chips[currchip]->sram_bytes - addrparse;
                }
				
                dump_sram(chips[currchip], addrparse, addrbytes);
                break;
				
            case 'U':
                /* Unassemble the flash and dump to a file */
                if(argc < 2){
                    unassemble(chips[currchip],currchip, NULL);
                } else {
                    unassemble(chips[currchip],currchip, argv[1]);
                }
                break;
				
            case 'B':
                /* Create a breakpoint */
				
                /* Check parameters */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
						" argument.\n", 'B');
                    continue;
                }
				
                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                /* Create the actual breakpoint */
                bpoint = (break_t*)malloc(sizeof(*bpoint));
                breakset++;
                bpoint->addr = addrparse / 2;
                log_debug(chips[currchip], V_QUIET,
					"breakpoint 0x%06x set.\n", addrparse);
                table_store(chips[currchip]->breaktable, bpoint);
                break;
				
            case 'W':
                /* Create a watchpoint */
				
                /* Check parameters */
                if(argc < 2) {
                    printf("Error: The %c command requires a numerical"
						" argument.\n", 'W');
                    continue;
                }
				
                /* Read the argument */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                if(argc == 3){
                    addrbytes = strtoul(argv[2], &lastchar, 16);
                    if((lastchar != NULL) && (*lastchar != 0)) {
                        /* Error: parameter contains illegal cahracters */
                        printf("Error: The value '%s' is not a valid"
							" hexadecimal number.\n", argv[1]);
                        continue;
                    }
                    if(addrbytes == 0 || addrparse + addrbytes > chips[currchip]->sram_bytes){
                        printf("Error: Watchpoint size must be >0 and fit within SRAM\n");
                        continue;
                    }
                }else{
                    addrbytes = 1;
                }
				
                /* Create the actual watchpoint */
                wpoint = (watch_t*)malloc(sizeof(*wpoint));
                watchset++;
                wpoint->addr = addrparse;
                wpoint->length = addrbytes;
                wpoint->value = malloc(addrbytes);
                log_debug(chips[currchip], V_QUIET,
					"watchpoint 0x%06x set.\n", addrparse);
                for(i=0; i < addrbytes; i++){
                    atmega_read_sram(chips[currchip], wpoint->addr + i, wpoint->value + i);
                }
                table_store(chips[currchip]->watchtable, wpoint);
                break;
				
            case 'N':
                /* Create a notify point */
				
                /* Check parameters */
                if(argc < 3) {
                    printf("Error: The %c command requires two numerical"
						" arguments.\n", 'N');
                    continue;
                }
				
                /* Read the notify address */
                addrparse = strtoul(argv[1], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[1]);
                    continue;
                }
				
                /* Read the notify identifier */
                intparse = strtol(argv[2], &lastchar, 0);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    printf("Error: The value '%s' is not a valid"
						" numerical value.\n", argv[2]);
                    continue;
                }
				
                /* Create the actual notifypoint */
                wpoint = (watch_t*)malloc(sizeof(*wpoint));
                wpoint->addr = addrparse;
                wpoint->id = intparse;
                wpoint->length = 1;
                wpoint->value = malloc(1);
                atmega_read_sram(chips[currchip], wpoint->addr,
					wpoint->value);
                table_store(chips[currchip]->notifytable, wpoint);
                break;
				
            case 'X':
                /* Delete a (break|watch|notify) point */
				
                /* Check parameters */
                if(argc < 3) {
                    printf("Error: The %c command requires a break point type"
						" and an address.\n", 'X');
                    continue;
                }
				
                /* Parse the address */
                addrparse = strtoul(argv[2], &lastchar, 16);
                if((lastchar != NULL) && (*lastchar != 0)) {
                    /* Error: parameter contains illegal cahracters */
                    printf("Error: The value '%s' is not a valid"
						" hexadecimal number.\n", argv[2]);
                    continue;
                }
				
                /* Handle the appropriate type of breakpoint */
                switch(toupper(argv[1][0])) {
                case 'B':
                    /* Break Point */
                    bquery.addr = addrparse / 2;
                    if(table_delete(chips[currchip]->breaktable, &bquery,
						free) >= 0) {
                        log_debug(chips[currchip], V_QUIET,
							"breakpoint 0x%06x cleared.\n", addrparse);
                        breakset--;
                    }else{
                        log_debug(chips[currchip], V_QUIET,
							"no such breakpoint.\n");
                    }
                    break;
					
                case 'W':
                    /* Watch Point */
                    wquery.addr = addrparse;
                    if(table_delete(chips[currchip]->watchtable, &wquery,
						free) >= 0){
                        log_debug(chips[currchip], V_QUIET,
							"watchpoint 0x%06x cleared.\n", addrparse);
                        watchset--;
                    }else{
                        log_debug(chips[currchip], V_QUIET,
							"no such watchpoint.\n");
                    }
                    break;
					
                case 'N':
                    /* Notify Point */
                    wquery.addr = addrparse;
                    if(table_delete(chips[currchip]->notifytable, &wquery,
						free) >= 0){
                        log_debug(chips[currchip], V_QUIET,
							"notifypoint cleared.\n");
                    }else{
                        log_debug(chips[currchip], V_QUIET,
							"no such notifypoint.\n");
                    }
                    
                    break;
					
                default:
                    printf("Error: Unknown breakpoint class '%s'.\n", argv[1]);
                    break;
                } /* switch(toupper(argv[1][0])) { */
				
                break;
				
				case 'G':
					/* Run until breakpoint */
					
					mainloop_go(chips, breakset, watchset);
					
					quit = 1;
					break;
					
				default:
					printf("Error: Unrecognized command '%s'\n", argv[0]);
            }
        }
    } /* while( !quit ) */
	
    free(argv);
    argv = NULL;
    argv_alloc = 0;
}

int main( int argc, char **argv )
{
    int cmderror = 0;
    error_t result;
    atmega_chip **chips;
    char *fname=NULL;
    char *config="mica2";
    char optstring[] = "htvqbc:l:VSTICMD";
    char option;
    int i, fd;
    static char startUART=0;
    int nLogs=0;
    
    printf("arg %s\n", argv[argc-1] );
    if(argv[argc-1][0]>'0'&&argv[argc-1][0]<='9')
		g_num_attributes= atoi(argv[argc-1]);
    if(g_num_attributes<1||g_num_attributes>3)
		g_num_attributes=1;
	//printf("g_num_attributes=%d\n",g_num_attributes);
    while( (option = getopt(argc, argv, optstring))!=EOF){
        switch (option){
		case '2':
            g_num_attributes=2;
			// printf("DDDDDDDDDDDDDDDDDDDDDDDDDDDDD\n");
            break;
		case '3':
            g_num_attributes=3;
            break;
            
		case 'c':
			/* single-run: machine config */
			if(optarg){
				config = optarg;
			}
			else{
				printf("option c requires a file name\n");
				cmderror=1;
			}
			break;
		case 'l': 
			//+++++++++++++++++++++++++++++++++++++++++
			//VMNet log power uses:
#ifndef POWER_CONSUMPTION_MEASURE
			if(optarg){
				fd = fork_compress(optarg);
				if (fd == -1) {
					cmderror = 1;
				} else {
					logfile = fdopen(fd, "w");
					if (logfile == NULL) {
						perror("Error opening write pipe file");
						cmderror = -1;
					}
				}
			} else {
				printf("option l requires a file name\n");
				cmderror = 1;
			}
#else
#endif                                   
			//++++++++++++++++++++++++++++++++++++++++++++++++++++
			break;
		case 'S':
			logstack = 1;
			break;
		case 'C':
			logcall = 1;
			break;
		case 'I':
			logintr = 1;
			break;
		case 'T':
			logtask = 1;
			break;
		case 'M':
			logmemory = 1;
			break;
		case 'D':
			logmachine = 1;
			srand48(time(NULL));
			break;
		case 'b':
			backend = 1;
			break;
		case 'h':
			printf("Options:\n");
			printf("\t-c <filename>\tspecify config file.  Default is mica2.conf\n");
			printf("\t-l <filename>\tspecify a log file.\n");
			printf("\t-C\t\tlog function calls, requires -l\n");
			printf("\t-S\t\tlog stack changes, requires -l\n");
			printf("\t-I\t\tlog interrupts and returns, requires -l\n");
			printf("\t-T\t\tlog TinyOS task info, requires -l\n");
			printf("\t-D\t\trandomly dump machine state info in log, requires -l\n");
			printf("\t-M\t\tlog memory operations, requires -l\n");
			printf("\t-t\t\trun in 'test mode'\n");
			printf("\t-b\t\trun in 'backend mode' (output altered for xatdb)\n");
			printf("\t-v\t\tincreases verbosity by one each time.\n");
			printf("\t-q\t\tdecreases verbosity by one each time.\n");
			printf("\t-V\t\tdisplay version information\n");
			printf("\t-h\t\tDisplay this list.\n\n");
			
			exit(0);
			break;
		case 'v':
			verbosity++;
			if(verbosity > V_ALL){
				verbosity=V_ALL;
			}
			break;
		case 'q':
			verbosity--;
			if(verbosity < V_QUIET){
				verbosity=V_QUIET;
			}
			break;
		case 'V':
			// printf( "[ATmega128L Emulator %s]\n", PACKAGE_VERSION );
			exit(0);
			break;
		case ':':
			printf("missing required parameter for an option\n");
			cmderror=1;
			break;
		case '?':
			printf("unknown option\n");
			cmderror=1;
			break;
		default:
			printf("Unknown getopt return code '%c'\n", option);
			cmderror=1;
			break;
        }
    }
	
    if(optind < argc){
        fname = argv[optind];
    }
    
    if(cmderror || (fname == NULL)){
        printf("Usage: %s [-options] filename\n", argv[0]);
        printf("\tUse -h for option list\n");
        exit(-1);
    }
    //++++++++++++++++++++++++++++++++++++++++++++++++++++
    //VMNet run the following way to ensure the instructions run forever.
    printf( "VMNet 1.0, Computer Science, HKUST [Copyright (c) 2004, HKUST]\n");
    printf( "Adopted some codes from Atemu: Copyright (c) 2003, University of Maryland \n  Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks \n");
    {
		g_idVMNet=0;
		
    }
    /* set up the chips */
	
    nchips = atemu_load_file( &chips, fname, config );
	
	g_nChips=nchips;
    g_chips=chips;
    
#ifdef LOG_INITIALIZATION_TIME    
    g_received_id= malloc(nchips*sizeof (uint32_t));
    g_epoch_id= malloc(nchips);
    if(!g_received_id||!g_epoch_id)
    {
        printf("Cannot malloc an array for chips initialization time\n");
        exit(-1);
    }
    memset(g_received_id,0,nchips);
    memset(g_epoch_id,0,(nchips*sizeof (uint32_t)));
#endif
	
#if 0
    g_current_waited_cycles=(int32_t*)malloc(nchips*sizeof (int32_t));
    if(!g_current_waited_cycles)
    {
        printf("Cannot malloc an array for chips waiting time\n");
        exit(-1);
    }
    memset(g_current_waited_cycles,0,nchips*sizeof (int32_t));
#endif
#ifdef POWER_CONSUMPTION_MEASURE    
    g_power_log=(FILE**)malloc(nchips*sizeof(FILE*));
    if(!g_power_log)
    {
        printf("Cannot malloc an array for log files\n");
        exit(-1);
    }
    while(nLogs<nchips)
    {
		char tempFileName[80]={0};
		sprintf(tempFileName,"logs/power_log_%d_0.rec",nLogs);
		g_power_log[nLogs]=fopen(tempFileName, "w");
		printf("log file name %s \n",tempFileName);
		nLogs++;
    }
#endif
    g_sample_data_count=(long*)malloc(nchips*sizeof(long));
    g_sensor_data=(unsigned short**)malloc(nchips*sizeof(unsigned short *));
    g_sensor_data_fptr=(FILE**)malloc(nchips*sizeof(FILE*));
    if(!g_sample_data_count||!g_sensor_data||!g_sensor_data_fptr)
    {
        printf("Cannot malloc memory...\n");
        exit(-1);
    }
    memset(g_sample_data_count,0,nchips*sizeof(long));
    memset(g_sensor_data,0,nchips*sizeof(unsigned short*));
    memset(g_sensor_data_fptr,0,nchips*sizeof(FILE*));
    nLogs=0;
    while(nLogs<nchips)
    {
        initiate_sensor_data(nLogs);
        nLogs++;
    }
	
	free(g_sensor_data_fptr);
	
    signal(SIGINT, signal_handler);
    signal(SIGCHLD, signal_handler);
    set_decode_mode(DECODE_NORMAL, NULL);
    
    //initializeRadioSockets(BIT_PORT+g_idVMNet);
	
#ifndef VMNET
    startValue=get_program_time(&startTime);
    g_isInitialized=1;
#endif
    if(!startUART){ 
		startUART=1;
		initializeSockets(); 
	}
	
    mainloop( chips );
    printf( "Done.\n" );
    stopValue=get_program_time(&stopTime);
    for(i=0; i < nchips; i++) {
		printf("%03d %20lld %s \n", chips[i]->id, chips[i]->cycles, "END");
    }
    printf("start time is %lld.%03lld us \n", startTime.tv_sec, startTime.tv_usec);    
    printf("End   time is %lld.%03lld us \n", stopTime.tv_sec, stopTime.tv_usec);    
    printf("Total time is %lld.%03lld us \n", (stopValue-startValue)/1000,(stopValue-startValue)%1000);
    for(i=0; i<nchips; i++){
        result = atmega_destroy_chip( &chips[i] );
        if ( result < 0 )
        {
            printf( "Error: Failed to destroy chip %d\n", i );
            return result;
        }
        printf( "Destroyed chip %d.\n", i );
    }  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	
    if(logfile != NULL) {
        fclose(logfile);
        close(g_fd);
    }
    printf( "Done.\n" );
	
    return 0;
}
int initiate_sensor_data(int n_idChip)
{
	
	int i,j;
	char tempFileName[80]={0} ;
	sprintf(tempFileName,"%s%d.dat","run/sensor",n_idChip);
	g_sensor_data_fptr[n_idChip]=fopen(tempFileName, "rb");
    // printf("g_sensor_data_fptr[n_idChip] =%x\n",g_sensor_data_fptr[n_idChip]);
	if(g_sensor_data_fptr[n_idChip])
	{
		int dataCounter =0;
		g_sensor_data[n_idChip]=malloc(MAX_SENSOR_DATA_SIZE*sizeof(unsigned short)*MAX_SENSOR_ATTRIBUTES);
		if(!g_sensor_data[n_idChip])
			return 1;
		while (!feof(g_sensor_data_fptr[n_idChip]))
		{
			i=0;j=0;
			memset(tempFileName,0,10);
			while((tempFileName[0]>'9'||tempFileName[0]<'0')&&!feof(g_sensor_data_fptr[n_idChip]))
                fread(&(tempFileName), 1, 1,  g_sensor_data_fptr[n_idChip]);
			while((tempFileName[i]<='9'&&tempFileName[i]>='0')&&!feof(g_sensor_data_fptr[n_idChip]))
			{
				i++;
				fread(&(tempFileName[i]), 1, 1,  g_sensor_data_fptr[n_idChip]);
			}
			tempFileName[i]='\0';
			if(i>1)
			{
				g_sensor_data[n_idChip][dataCounter++] = atoi(tempFileName);
				if(dataCounter>=(MAX_SENSOR_DATA_SIZE*MAX_SENSOR_ATTRIBUTES))
					break;
			}
		}
		fclose(g_sensor_data_fptr[n_idChip]);
	}
	return 0;
}

